from collections.abc import AsyncIterator, Generator, Iterator
from unittest.mock import AsyncMock, MagicMock

import httpx
import pytest

from httpx_retries import Retry, RetryTransport, aretry_request, retry_request


class _FailingByteStream(httpx.SyncByteStream):
    """A response body that raises when read."""

    def __init__(self, exc: Exception) -> None:
        self._exc = exc

    def __iter__(self) -> Iterator[bytes]:
        raise self._exc
        yield b""  # pragma: no cover

    def close(self) -> None:
        pass


class _FailingAsyncByteStream(httpx.AsyncByteStream):
    """A response body that raises when read."""

    def __init__(self, exc: Exception) -> None:
        self._exc = exc

    async def __aiter__(self) -> AsyncIterator[bytes]:
        raise self._exc
        yield b""  # pragma: no cover

    async def aclose(self) -> None:
        pass


class BodyFailTransport(httpx.BaseTransport):
    """Returns a response whose body read fails `fail_times` times, then succeeds."""

    def __init__(self, exc: Exception, fail_times: int) -> None:
        self.exc = exc
        self.fail_times = fail_times
        self.attempts = 0

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        self.attempts += 1
        if self.attempts <= self.fail_times:
            return httpx.Response(200, stream=_FailingByteStream(self.exc))
        return httpx.Response(200, content=b"ok")


class AsyncBodyFailTransport(httpx.AsyncBaseTransport):
    """Returns a response whose body read fails `fail_times` times, then succeeds."""

    def __init__(self, exc: Exception, fail_times: int) -> None:
        self.exc = exc
        self.fail_times = fail_times
        self.attempts = 0

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        self.attempts += 1
        if self.attempts <= self.fail_times:
            return httpx.Response(200, stream=_FailingAsyncByteStream(self.exc))
        return httpx.Response(200, content=b"ok")


class _HeaderAuth(httpx.Auth):
    """Auth scheme that stamps a fixed header onto each request."""

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        request.headers["X-Auth"] = "secret"
        yield request


class RecordingTransport(httpx.BaseTransport):
    """Records the requests it receives and returns 200."""

    def __init__(self) -> None:
        self.requests: list[httpx.Request] = []

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        self.requests.append(request)
        return httpx.Response(200, content=b"ok")


class AsyncRecordingTransport(httpx.AsyncBaseTransport):
    """Records the requests it receives and returns 200."""

    def __init__(self) -> None:
        self.requests: list[httpx.Request] = []

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        self.requests.append(request)
        return httpx.Response(200, content=b"ok")


class RedirectTransport(httpx.BaseTransport):
    """Redirects `/start` to `/final` once, then serves `/final`."""

    def __init__(self) -> None:
        self.paths: list[str] = []

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        self.paths.append(request.url.path)
        if request.url.path == "/start":
            return httpx.Response(302, headers={"Location": "https://example.com/final"})
        return httpx.Response(200, content=b"final")


class AsyncRedirectTransport(httpx.AsyncBaseTransport):
    """Redirects `/start` to `/final` once, then serves `/final`."""

    def __init__(self) -> None:
        self.paths: list[str] = []

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        self.paths.append(request.url.path)
        if request.url.path == "/start":
            return httpx.Response(302, headers={"Location": "https://example.com/final"})
        return httpx.Response(200, content=b"final")


def test_retry_request_success(mock_sleep: MagicMock) -> None:
    transport = BodyFailTransport(httpx.ReadTimeout("boom"), fail_times=0)

    with httpx.Client(transport=transport) as client:
        response = retry_request(client, "GET", "https://example.com")

    assert response.status_code == 200
    assert response.text == "ok"
    assert mock_sleep.call_count == 0
    assert response.extensions["retry"].attempts_made == 0


# Both are in the default retryable set and are the two headline body-phase failures.
BODY_READ_ERRORS = [
    httpx.ReadTimeout("read timed out"),
    httpx.RemoteProtocolError("peer closed connection without sending complete message body"),
]


@pytest.mark.parametrize("exc", BODY_READ_ERRORS)
def test_retry_request_retries_body_read_error(mock_sleep: MagicMock, exc: Exception) -> None:
    transport = BodyFailTransport(exc, fail_times=2)

    with httpx.Client(transport=transport) as client:
        response = retry_request(client, "GET", "https://example.com")

    assert response.status_code == 200
    assert response.text == "ok"
    assert transport.attempts == 3
    assert mock_sleep.call_count == 2
    assert response.extensions["retry"].attempts_made == 2


def test_retry_request_raises_when_exhausted(mock_sleep: MagicMock) -> None:
    transport = BodyFailTransport(httpx.ReadTimeout("boom"), fail_times=100)
    retry = Retry(total=3)

    with httpx.Client(transport=transport) as client:
        with pytest.raises(httpx.ReadTimeout, match="boom"):
            retry_request(client, "GET", "https://example.com", retry=retry)

    assert mock_sleep.call_count == 3


def test_retry_request_non_retryable_method(mock_sleep: MagicMock) -> None:
    transport = BodyFailTransport(httpx.ReadTimeout("boom"), fail_times=1)

    with httpx.Client(transport=transport) as client:
        with pytest.raises(httpx.ReadTimeout, match="boom"):
            retry_request(client, "POST", "https://example.com")

    assert transport.attempts == 1
    assert mock_sleep.call_count == 0


def test_retry_request_extension_overrides_retry_argument(mock_sleep: MagicMock) -> None:
    transport = BodyFailTransport(httpx.ReadTimeout("boom"), fail_times=100)

    with httpx.Client(transport=transport) as client:
        with pytest.raises(httpx.ReadTimeout, match="boom"):
            retry_request(
                client,
                "GET",
                "https://example.com",
                retry=Retry(total=10),
                extensions={"retry": Retry(total=2)},
            )

    assert mock_sleep.call_count == 2


def test_retry_request_async_validate_response_raises_for_sync_client() -> None:
    async def validate(response: httpx.Response) -> None:  # pragma: no cover
        pass

    transport = BodyFailTransport(httpx.ReadTimeout("boom"), fail_times=0)
    retry = Retry(validate_response=validate)

    with httpx.Client(transport=transport) as client:
        with pytest.raises(TypeError, match="validate_response must be a sync function"):
            retry_request(client, "GET", "https://example.com", retry=retry)


@pytest.mark.asyncio
async def test_aretry_request_success(mock_asleep: AsyncMock) -> None:
    transport = AsyncBodyFailTransport(httpx.ReadTimeout("boom"), fail_times=0)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await aretry_request(client, "GET", "https://example.com")

    assert response.status_code == 200
    assert response.text == "ok"
    assert mock_asleep.call_count == 0


@pytest.mark.parametrize("exc", BODY_READ_ERRORS)
@pytest.mark.asyncio
async def test_aretry_request_retries_body_read_error(mock_asleep: AsyncMock, exc: Exception) -> None:
    transport = AsyncBodyFailTransport(exc, fail_times=2)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await aretry_request(client, "GET", "https://example.com")

    assert response.status_code == 200
    assert response.text == "ok"
    assert transport.attempts == 3
    assert mock_asleep.call_count == 2
    assert response.extensions["retry"].attempts_made == 2


@pytest.mark.asyncio
async def test_aretry_request_non_retryable_method(mock_asleep: AsyncMock) -> None:
    transport = AsyncBodyFailTransport(httpx.ReadTimeout("boom"), fail_times=1)

    async with httpx.AsyncClient(transport=transport) as client:
        with pytest.raises(httpx.ReadTimeout, match="boom"):
            await aretry_request(client, "POST", "https://example.com")

    assert transport.attempts == 1
    assert mock_asleep.call_count == 0


def test_retry_request_forwards_auth(mock_sleep: MagicMock) -> None:
    transport = RecordingTransport()

    with httpx.Client(transport=transport) as client:
        response = retry_request(client, "GET", "https://example.com", auth=_HeaderAuth())

    assert response.status_code == 200
    assert transport.requests[0].headers["X-Auth"] == "secret"


def test_retry_request_forwards_follow_redirects(mock_sleep: MagicMock) -> None:
    transport = RedirectTransport()

    with httpx.Client(transport=transport) as client:
        # Without follow_redirects, the client default (off) leaves the 302 in place.
        response = retry_request(client, "GET", "https://example.com/start")
        assert response.status_code == 302

        response = retry_request(client, "GET", "https://example.com/start", follow_redirects=True)

    assert response.status_code == 200
    assert response.text == "final"
    assert transport.paths == ["/start", "/start", "/final"]


@pytest.mark.asyncio
async def test_aretry_request_forwards_auth(mock_asleep: AsyncMock) -> None:
    transport = AsyncRecordingTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await aretry_request(client, "GET", "https://example.com", auth=_HeaderAuth())

    assert response.status_code == 200
    assert transport.requests[0].headers["X-Auth"] == "secret"


@pytest.mark.asyncio
async def test_aretry_request_forwards_follow_redirects(mock_asleep: AsyncMock) -> None:
    transport = AsyncRedirectTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await aretry_request(client, "GET", "https://example.com/start", follow_redirects=True)

    assert response.status_code == 200
    assert response.text == "final"
    assert transport.paths == ["/start", "/final"]


def test_retry_request_rejects_retrying_client() -> None:
    with httpx.Client(transport=RetryTransport(transport=RecordingTransport())) as client:
        with pytest.raises(ValueError, match="would retry every request twice"):
            retry_request(client, "GET", "https://example.com")


def test_retry_request_rejects_mounted_retrying_transport() -> None:
    mounts = {"https://": RetryTransport(transport=RecordingTransport())}
    with httpx.Client(mounts=mounts, trust_env=False) as client:
        with pytest.raises(ValueError, match="would retry every request twice"):
            retry_request(client, "GET", "https://example.com")


@pytest.mark.asyncio
async def test_aretry_request_rejects_retrying_client() -> None:
    async with httpx.AsyncClient(transport=RetryTransport(transport=AsyncRecordingTransport())) as client:
        with pytest.raises(ValueError, match="would retry every request twice"):
            await aretry_request(client, "GET", "https://example.com")


@pytest.mark.asyncio
async def test_aretry_request_rejects_mounted_retrying_transport() -> None:
    mounts = {"https://": RetryTransport(transport=AsyncRecordingTransport())}
    async with httpx.AsyncClient(mounts=mounts, trust_env=False) as client:
        with pytest.raises(ValueError, match="would retry every request twice"):
            await aretry_request(client, "GET", "https://example.com")


def test_retry_request_validate_response_retries(mock_sleep: MagicMock) -> None:
    transport = RecordingTransport()
    statuses: list[int] = []

    def validate(response: httpx.Response) -> None:
        statuses.append(response.status_code)
        if len(statuses) < 3:
            raise httpx.TimeoutException("not ready yet")

    retry = Retry(total=5, validate_response=validate)

    with httpx.Client(transport=transport) as client:
        response = retry_request(client, "GET", "https://example.com", retry=retry)

    assert response.status_code == 200
    assert statuses == [200, 200, 200]
    assert mock_sleep.call_count == 2



from unittest.mock import AsyncMock, MagicMock

import pytest


@pytest.fixture
def mock_sleep(monkeypatch: pytest.MonkeyPatch) -> MagicMock:
    mock_sleep = MagicMock()
    monkeypatch.setattr("time.sleep", mock_sleep)
    return mock_sleep


@pytest.fixture
def mock_asleep(monkeypatch: pytest.MonkeyPatch) -> MagicMock:
    mock_asleep = AsyncMock()
    monkeypatch.setattr("asyncio.sleep", mock_asleep)
    return mock_asleep



import asyncio
import logging
import time
from collections.abc import AsyncGenerator, Generator
from unittest.mock import AsyncMock, MagicMock, Mock, call, patch

import httpx
import pytest
from httpx import Request, Response

from httpx_retries import Retry, RetryTransport
from httpx_retries.transport import _retry_operation, _retry_operation_async


def status_codes(
    codes: list[tuple[int, str | None]],
) -> Generator[tuple[int, str | None], None, None]:
    """Yields the given status codes, and then the last status code indefinitely."""
    yield from codes
    while True:
        yield codes[-1]


async def astatus_codes(
    codes: list[tuple[int, str | None]],
) -> AsyncGenerator[tuple[int, str | None], None]:
    """Yields the given status codes, and then the last status code indefinitely."""
    for code in codes:
        yield code

    while True:
        yield codes[-1]


def create_response(request: Request, status_code: int, retry_after: str | None = None) -> Response:
    """Helper to create a response with the given status code and retry-after header"""
    headers = {"Retry-After": retry_after} if retry_after else {}
    return Response(status_code=status_code, request=request, headers=headers)


StatusCodeTuple = tuple[int, str | None]
StatusCodeSequence = Generator[StatusCodeTuple, None, None]
AsyncStatusCodeSequence = AsyncGenerator[StatusCodeTuple, None]
MockResponse = tuple[MagicMock, dict[str, StatusCodeSequence | None]]
AsyncMockResponse = tuple[AsyncMock, dict[str, AsyncStatusCodeSequence | None]]


@pytest.fixture
def mock_responses(mock_sleep: MagicMock) -> Generator[MockResponse, None, None]:
    """Returns a mock for sleep and response sequences for sync requests"""
    status_code_sequences: dict[str, StatusCodeSequence | None] = {}

    def handle_request(request: Request) -> Response:
        if request.url in status_code_sequences:
            status_code_gen = status_code_sequences[str(request.url)]
            if status_code_gen is not None:
                status_code, retry_after = next(status_code_gen)
                return create_response(request, status_code, retry_after)
        return create_response(request, 200)

    with patch("httpx.HTTPTransport.handle_request") as mock_handle:
        mock_handle.side_effect = handle_request
        yield mock_sleep, status_code_sequences


@pytest.fixture
def mock_async_responses(
    mock_asleep: AsyncMock,
) -> Generator[AsyncMockResponse, None, None]:
    """Returns a mock for sleep and response sequences for async requests"""
    status_code_sequences: dict[str, AsyncStatusCodeSequence | None] = {}

    async def handle_async_request(request: Request) -> Response:
        if request.url in status_code_sequences:
            status_code_gen = status_code_sequences[str(request.url)]
            if status_code_gen is not None:
                status_code, retry_after = await status_code_gen.__anext__()
                return create_response(request, status_code, retry_after)
        return create_response(request, 200)

    with patch("httpx.AsyncHTTPTransport.handle_async_request") as mock_handle:
        mock_handle.side_effect = handle_async_request
        yield mock_asleep, status_code_sequences


class MockHTTPTransport(httpx.HTTPTransport):
    def handle_request(self, request: Request) -> Response:
        return create_response(request, 200)


class MockAsyncHTTPTransport(httpx.AsyncHTTPTransport):
    async def handle_async_request(self, request: Request) -> Response:
        return create_response(request, 200)


class MockTransport(httpx.BaseTransport):
    def handle_request(self, request: Request) -> Response:
        return create_response(request, 200)


class MockAsyncTransport(httpx.AsyncBaseTransport):
    async def handle_async_request(self, request: Request) -> Response:
        return create_response(request, 200)


def test_successful_request(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com")

    assert response.status_code == 200
    assert mock_sleep.call_count == 0


def test_successful_request_logs(mock_responses: MockResponse, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG)
    mock_sleep, _ = mock_responses
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com")

    assert response.status_code == 200
    assert mock_sleep.call_count == 0
    assert "handle_request started request=<Request('GET', 'https://example.com')>" in caplog.text
    assert (
        "handle_request finished request=<Request('GET', 'https://example.com')> response=<Response [200 OK]>"
        in caplog.text
    )


def test_failed_request(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, None)])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")

    assert response.status_code == 429
    assert mock_sleep.call_count == 10


def test_unretryable_status_code(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(403, None), (200, None)])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")
        assert response.status_code == 403

    assert mock_sleep.call_count == 0


def test_unretryable_method(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, None), (200, None)])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.post("https://example.com/fail")
        assert response.status_code == 429

    assert mock_sleep.call_count == 0


def test_non_standard_method_passes_through(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/dav"] = status_codes([(207, None)])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.request("PROPFIND", "https://example.com/dav")
        assert response.status_code == 207

    assert mock_sleep.call_count == 0


@pytest.mark.asyncio
async def test_async_non_standard_method_passes_through(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, status_code_sequences = mock_async_responses
    status_code_sequences["https://example.com/dav"] = astatus_codes([(207, None)])
    transport = RetryTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.request("PROPFIND", "https://example.com/dav")
        assert response.status_code == 207

    assert mock_asleep.call_count == 0


def test_unretryable_exception(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses
    transport = RetryTransport()

    with patch(
        "httpx.HTTPTransport.handle_request",
        side_effect=httpx.ProxyError("Proxy error"),
    ):
        with httpx.Client(transport=transport) as client:
            with pytest.raises(httpx.ProxyError, match="Proxy error"):
                client.get("https://example.com")

    assert mock_sleep.call_count == 0


@pytest.mark.asyncio
async def test_async_unretryable_exception(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, _ = mock_async_responses
    transport = RetryTransport()

    with patch(
        "httpx.AsyncHTTPTransport.handle_async_request",
        side_effect=httpx.ProxyError("Proxy error"),
    ):
        async with httpx.AsyncClient(transport=transport) as client:
            with pytest.raises(httpx.ProxyError, match="Proxy error"):
                await client.get("https://example.com")

    assert mock_asleep.call_count == 0


def test_retryable_exception(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses
    transport = RetryTransport()

    with patch("httpx.HTTPTransport.handle_request", side_effect=httpx.ReadTimeout("Timeout!")):
        with httpx.Client(transport=transport) as client:
            with pytest.raises(httpx.ReadTimeout, match="Timeout!"):
                client.get("https://example.com")

    assert mock_sleep.call_count == 10


def test_retryable_exception_custom_exception(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses
    transport = RetryTransport(retry=Retry(retry_on_exceptions=[ValueError]))

    with patch("httpx.HTTPTransport.handle_request", side_effect=ValueError("oops")):
        with httpx.Client(transport=transport) as client:
            with pytest.raises(ValueError, match="oops"):
                client.get("https://example.com")

    assert mock_sleep.call_count == 10


@pytest.mark.parametrize("status_code", Retry.RETRYABLE_STATUS_CODES)
def test_retry_operation_always_closes_response(status_code: int) -> None:
    responses = []

    def send_method(request: httpx.Request) -> httpx.Response:
        response = Mock(spec=httpx.Response)
        response.status_code = status_code
        response.headers = httpx.Headers()
        response.extensions = {}
        response.close = Mock()

        responses.append(response)
        return response

    _retry_operation(request=httpx.Request("GET", "https://example.com"), send_method=send_method, retry=Retry())

    assert all(r.close.called for r in responses[:-1])


@pytest.mark.asyncio
async def test_async_retryable_exception(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, _ = mock_async_responses
    transport = RetryTransport()

    with patch(
        "httpx.AsyncHTTPTransport.handle_async_request",
        side_effect=httpx.ReadTimeout("oops"),
    ):
        async with httpx.AsyncClient(transport=transport) as client:
            with pytest.raises(httpx.ReadTimeout, match="oops"):
                await client.get("https://example.com")

    assert mock_asleep.call_count == 10


@pytest.mark.asyncio
async def test_async_retryable_exception_custom_exception(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, _ = mock_async_responses
    transport = RetryTransport(retry=Retry(retry_on_exceptions=[ValueError]))

    with patch(
        "httpx.AsyncHTTPTransport.handle_async_request",
        side_effect=ValueError("Timeout!"),
    ):
        async with httpx.AsyncClient(transport=transport) as client:
            with pytest.raises(ValueError, match="Timeout!"):
                await client.get("https://example.com")

    assert mock_asleep.call_count == 10


@pytest.mark.asyncio
async def test_successful_async_request_logs(
    mock_async_responses: AsyncMockResponse, caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.DEBUG)
    mock_asleep, _ = mock_async_responses
    transport = RetryTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")

    assert response.status_code == 200
    assert mock_asleep.call_count == 0
    assert "handle_async_request started request=<Request('GET', 'https://example.com')>" in caplog.text
    assert (
        "handle_async_request finished request=<Request('GET', 'https://example.com')> response=<Response [200 OK]>"
        in caplog.text
    )


def test_custom_retryable_exception(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses

    retry = Retry(retry_on_exceptions=[httpx.ProxyError])
    transport = RetryTransport(retry=retry)

    with patch(
        "httpx.HTTPTransport.handle_request",
        side_effect=httpx.ProxyError("Proxy error"),
    ):
        with httpx.Client(transport=transport) as client:
            with pytest.raises(httpx.ProxyError, match="Proxy error"):
                client.get("https://example.com")

    assert mock_sleep.call_count == 10

    # Verify other exceptions are not retried
    transport = RetryTransport(retry=retry)
    with patch("httpx.HTTPTransport.handle_request", side_effect=httpx.ReadTimeout("Timeout!")):
        with httpx.Client(transport=transport) as client:
            with pytest.raises(httpx.ReadTimeout, match="Timeout!"):
                client.get("https://example.com")

    assert mock_sleep.call_count == 10  # Count shouldn't increase


def test_retries_reset_for_new_request(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, None)])
    status_code_sequences["https://example.com/fail2"] = status_codes([(429, None)])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")
        assert response.status_code == 429

        response = client.get("https://example.com/fail2")
        assert response.status_code == 429

    assert mock_sleep.call_count == 20


def test_retry_respects_retry_after_header(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, "5")])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")
        assert response.status_code == 429

    assert mock_sleep.call_count == 10
    mock_sleep.assert_has_calls([call(5)] * 10)


def test_transport_logs_retry_operation(mock_responses: MockResponse, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG)
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, "5")])
    transport = RetryTransport()

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")
        assert response.status_code == 429

    records = [r for r in caplog.records if r.message.startswith("_retry_operation")]

    assert len(records) == 10
    assert records[0].message == (
        "_retry_operation retrying request=<Request('GET', 'https://example.com/fail')> "
        "response=<Response [429 Too Many Requests]> retry=<Retry(total=10, attempts_made=0)>"
    )
    assert records[-1].message == (
        "_retry_operation retrying request=<Request('GET', 'https://example.com/fail')> "
        "response=<Response [429 Too Many Requests]> retry=<Retry(total=10, attempts_made=9)>"
    )


@pytest.mark.asyncio
async def test_async_retry_operation_logs(
    mock_async_responses: AsyncMockResponse, caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level(logging.DEBUG)
    mock_asleep, status_code_sequences = mock_async_responses
    status_code_sequences["https://example.com/fail"] = astatus_codes([(429, "5")])
    transport = RetryTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com/fail")
        assert response.status_code == 429

    records = [r for r in caplog.records if r.message.startswith("_retry_operation_async")]
    assert len(records) == 10
    assert records[0].message == (
        "_retry_operation_async retrying request=<Request('GET', 'https://example.com/fail')> "
        "response=<Response [429 Too Many Requests]> retry=<Retry(total=10, attempts_made=0)>"
    )
    assert records[-1].message == (
        "_retry_operation_async retrying request=<Request('GET', 'https://example.com/fail')> "
        "response=<Response [429 Too Many Requests]> retry=<Retry(total=10, attempts_made=9)>"
    )


@pytest.mark.asyncio
async def test_async_successful_request(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, status_code_sequences = mock_async_responses
    transport = RetryTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")

    assert response.status_code == 200
    assert mock_asleep.call_count == 0


@pytest.mark.asyncio
async def test_async_failed_request(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, status_code_sequences = mock_async_responses
    status_code_sequences["https://example.com/fail"] = astatus_codes([(429, None)])
    transport = RetryTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com/fail")

    assert response.status_code == 429
    assert mock_asleep.call_count == 10


@pytest.mark.asyncio
async def test_sync_only_transport() -> None:
    transport = RetryTransport(transport=MockHTTPTransport())

    # Sync works
    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com")
        assert response.status_code == 200

    # Async fails
    with pytest.raises(RuntimeError, match="Async request received but no async transport available"):
        async with httpx.AsyncClient(transport=transport) as client:
            await client.get("https://example.com")


@pytest.mark.asyncio
async def test_async_only_transport() -> None:
    transport = RetryTransport(transport=MockAsyncHTTPTransport())

    # Async works
    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")
        assert response.status_code == 200

    # Sync fails
    with pytest.raises(
        RuntimeError,
        match="Synchronous request received but no sync transport available",
    ):
        with httpx.Client(transport=transport) as client:
            client.get("https://example.com")


@pytest.mark.asyncio
async def test_async_unretryable_method(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, status_code_sequences = mock_async_responses
    status_code_sequences["https://example.com/fail"] = astatus_codes([(429, None), (200, None)])
    transport = RetryTransport()

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.post("https://example.com/fail")
        assert response.status_code == 429

    assert mock_asleep.call_count == 0


@pytest.mark.asyncio
async def test_sync_from_base_transport() -> None:
    transport = RetryTransport(transport=MockTransport())

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com")
        assert response.status_code == 200


@pytest.mark.asyncio
async def test_async_from_base_transport() -> None:
    transport = RetryTransport(transport=MockAsyncTransport())

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")
        assert response.status_code == 200


def test_retry_extension_overrides_transport(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, None)])
    transport = RetryTransport(retry=Retry(total=10))

    request = httpx.Request("GET", "https://example.com/fail", extensions={"retry": Retry(total=2)})
    with httpx.Client(transport=transport) as client:
        response = client.send(request)

    assert response.status_code == 429
    assert mock_sleep.call_count == 2
    assert response.extensions["retry"].attempts_made == 2


def test_retry_extension_set_from_transport_when_absent(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses
    transport = RetryTransport(retry=Retry(total=3))

    request = httpx.Request("GET", "https://example.com")
    with httpx.Client(transport=transport) as client:
        response = client.send(request)

    assert request.extensions["retry"] is transport.retry
    assert response.extensions["retry"] is transport.retry


@pytest.mark.asyncio
async def test_async_retry_extension_overrides_transport(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, status_code_sequences = mock_async_responses
    status_code_sequences["https://example.com/fail"] = astatus_codes([(429, None)])
    transport = RetryTransport(retry=Retry(total=10))

    request = httpx.Request("GET", "https://example.com/fail", extensions={"retry": Retry(total=2)})
    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.send(request)

    assert response.status_code == 429
    assert mock_asleep.call_count == 2
    assert response.extensions["retry"].attempts_made == 2


@pytest.mark.asyncio
async def test_async_retry_extension_set_from_transport_when_absent(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, _ = mock_async_responses
    transport = RetryTransport(retry=Retry(total=3))

    request = httpx.Request("GET", "https://example.com")
    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.send(request)

    assert request.extensions["retry"] is transport.retry
    assert response.extensions["retry"] is transport.retry


def test_retry_after_capped_by_total_timeout(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(429, "120")])
    retry = Retry(total=10, total_timeout=10)
    transport = RetryTransport(retry=retry)

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")

    assert response.status_code == 429
    total_slept = sum(c.args[0] for c in mock_sleep.call_args_list)
    assert total_slept <= 10
    assert mock_sleep.call_count < 10


@pytest.mark.asyncio
async def test_async_retry_after_capped_by_total_timeout(
    mock_async_responses: AsyncMockResponse,
) -> None:
    mock_asleep, status_code_sequences = mock_async_responses
    status_code_sequences["https://example.com/fail"] = astatus_codes([(429, "120")])
    retry = Retry(total=10, total_timeout=10)
    transport = RetryTransport(retry=retry)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com/fail")

    assert response.status_code == 429
    total_slept = sum(c.args[0] for c in mock_asleep.call_args_list)
    assert total_slept <= 10
    assert mock_asleep.call_count < 10


@pytest.mark.parametrize("status_code", Retry.RETRYABLE_STATUS_CODES)
@pytest.mark.asyncio
async def test_retry_operation_async_always_closes_response(status_code: int) -> None:
    responses = []

    async def send_method(request: httpx.Request) -> httpx.Response:
        response = AsyncMock(spec=httpx.Response)
        response.status_code = status_code
        response.headers = httpx.Headers()
        response.extensions = {}
        response.aclose = AsyncMock()

        responses.append(response)
        return response

    await _retry_operation_async(
        request=httpx.Request("GET", "https://example.com"), send_method=send_method, retry=Retry()
    )

    assert all(r.aclose.called for r in responses[:-1])


# a retry backoff on one coroutine must not block peer coroutines sharing the
# event loop. These tests intentionally avoid the mock_asleep fixture because
# mocking asyncio.sleep short-circuits the scheduler and defeats the check.


@pytest.mark.asyncio
async def test_async_retry_does_not_block_peer_coroutine() -> None:
    slow_calls = 0

    async def handle_request(request: Request) -> Response:
        nonlocal slow_calls
        if str(request.url) == "https://example.com/slow":
            slow_calls += 1
            if slow_calls == 1:
                return create_response(request, 503)
        return create_response(request, 200)

    with patch("httpx.AsyncHTTPTransport.handle_async_request", side_effect=handle_request):
        retry = Retry(total=3, backoff_factor=0.2, backoff_jitter=0.0)
        transport = RetryTransport(retry=retry)

        async with httpx.AsyncClient(transport=transport) as client:
            slow_task = asyncio.create_task(client.get("https://example.com/slow"))

            # Let slow reach its first 503 and enter asyncio.sleep before we race.
            await asyncio.sleep(0.05)

            fast_start = time.monotonic()
            fast_response = await client.get("https://example.com/fast")
            fast_elapsed = time.monotonic() - fast_start

            slow_response = await slow_task

        assert fast_response.status_code == 200
        assert slow_response.status_code == 200
        # Backoff for attempts_made=1 is 0.2 * 2 = 0.4s; fast must not be serialised behind it.
        assert fast_elapsed < 0.2, f"fast request took {fast_elapsed:.3f}s — retry appears to block peers"


@pytest.mark.asyncio
async def test_async_concurrent_retries_do_not_serialize() -> None:
    call_counts: dict[str, int] = {}

    async def handle_request(request: Request) -> Response:
        url = str(request.url)
        call_counts[url] = call_counts.get(url, 0) + 1
        if call_counts[url] == 1:
            return create_response(request, 503)
        return create_response(request, 200)

    with patch("httpx.AsyncHTTPTransport.handle_async_request", side_effect=handle_request):
        retry = Retry(total=2, backoff_factor=0.15, backoff_jitter=0.0)
        transport = RetryTransport(retry=retry)

        async with httpx.AsyncClient(transport=transport) as client:
            start = time.monotonic()
            responses = await asyncio.gather(
                client.get("https://example.com/a"),
                client.get("https://example.com/b"),
                client.get("https://example.com/c"),
            )
            elapsed = time.monotonic() - start

        assert [r.status_code for r in responses] == [200, 200, 200]
        # Each request sleeps 0.3s once. Parallel ≈ 0.3s; serialised ≈ 0.9s.
        assert elapsed < 0.7, f"three concurrent retries took {elapsed:.3f}s — they look serialised"


@pytest.mark.asyncio
async def test_async_retry_sleep_yields_to_event_loop() -> None:
    async def handle_request(request: Request) -> Response:
        if str(request.url) == "https://example.com/retry":
            if handle_request.calls == 0:  # type: ignore[attr-defined]
                handle_request.calls += 1  # type: ignore[attr-defined]
                return create_response(request, 503)
        return create_response(request, 200)

    handle_request.calls = 0  # type: ignore[attr-defined]

    counter = 0
    stop = asyncio.Event()

    async def ticker() -> None:
        nonlocal counter
        while not stop.is_set():
            await asyncio.sleep(0.01)
            counter += 1

    with patch("httpx.AsyncHTTPTransport.handle_async_request", side_effect=handle_request):
        retry = Retry(total=3, backoff_factor=0.2, backoff_jitter=0.0)
        transport = RetryTransport(retry=retry)

        async with httpx.AsyncClient(transport=transport) as client:
            ticker_task = asyncio.create_task(ticker())
            response = await client.get("https://example.com/retry")
            stop.set()
            await ticker_task

    assert response.status_code == 200
    # Retry sleep is 0.4s. A non-blocked loop should tick ~40 times at 10ms cadence;
    # a blocked loop would tick 0 or 1 times.
    assert counter > 5, f"ticker only advanced {counter} times during retry — event loop looks blocked"


@pytest.mark.asyncio
async def test_async_shared_transport_isolates_retry_state() -> None:
    call_counts: dict[str, int] = {}

    async def handle_request(request: Request) -> Response:
        url = str(request.url)
        call_counts[url] = call_counts.get(url, 0) + 1
        if call_counts[url] <= 2:
            return create_response(request, 503)
        return create_response(request, 200)

    with patch("httpx.AsyncHTTPTransport.handle_async_request", side_effect=handle_request):
        retry = Retry(total=3, total_timeout=10.0, backoff_factor=0.05, backoff_jitter=0.0)
        transport = RetryTransport(retry=retry)

        async with httpx.AsyncClient(transport=transport) as client:
            responses = await asyncio.gather(
                client.get("https://example.com/x"),
                client.get("https://example.com/y"),
            )

        assert [r.status_code for r in responses] == [200, 200]
        # The transport's seed Retry must remain pristine — concurrent requests
        # must not mutate it, or retry budgets leak across requests.
        assert transport.retry.attempts_made == 0
        assert transport.retry.elapsed_sleep == 0.0


def test_validate_response_retries_on_failure(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    call_count = 0

    def validate(response: httpx.Response) -> None:
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise httpx.TimeoutException("not ready yet")

    retry = Retry(total=5, validate_response=validate)
    transport = RetryTransport(retry=retry)

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com")

    assert response.status_code == 200
    assert call_count == 3
    assert mock_sleep.call_count == 2


def test_validate_response_non_retryable_exception_raises(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses

    def validate(response: httpx.Response) -> None:
        raise ValueError("bad response")

    retry = Retry(total=5, validate_response=validate)
    transport = RetryTransport(retry=retry)

    with pytest.raises(ValueError, match="bad response"):
        with httpx.Client(transport=transport) as client:
            client.get("https://example.com")

    assert mock_sleep.call_count == 0


def test_validate_response_exhausted_returns_response(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses

    def validate(response: httpx.Response) -> None:
        raise httpx.TimeoutException("always bad")

    retry = Retry(total=3, validate_response=validate)
    transport = RetryTransport(retry=retry)

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com")

    assert response.status_code == 200
    assert mock_sleep.call_count == 3


def test_validate_response_async_callback_raises_for_sync_transport(mock_responses: MockResponse) -> None:
    mock_sleep, _ = mock_responses

    async def validate(response: httpx.Response) -> None:  # pragma: no cover
        pass

    retry = Retry(total=3, validate_response=validate)
    transport = RetryTransport(retry=retry)

    with pytest.raises(TypeError, match="validate_response must be a sync function"):
        with httpx.Client(transport=transport) as client:
            client.get("https://example.com")


@pytest.mark.asyncio
async def test_async_validate_response_retries_on_failure(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, _ = mock_async_responses
    call_count = 0

    async def validate(response: httpx.Response) -> None:
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise httpx.TimeoutException("not ready yet")

    retry = Retry(total=5, validate_response=validate)
    transport = RetryTransport(retry=retry)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")

    assert response.status_code == 200
    assert call_count == 3
    assert mock_asleep.call_count == 2


@pytest.mark.asyncio
async def test_async_validate_response_non_retryable_exception_raises(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, _ = mock_async_responses

    async def validate(response: httpx.Response) -> None:
        raise ValueError("bad response")

    retry = Retry(total=5, validate_response=validate)
    transport = RetryTransport(retry=retry)

    with pytest.raises(ValueError, match="bad response"):
        async with httpx.AsyncClient(transport=transport) as client:
            await client.get("https://example.com")

    assert mock_asleep.call_count == 0


@pytest.mark.asyncio
async def test_async_validate_response_sync_callback(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, _ = mock_async_responses
    call_count = 0

    def validate(response: httpx.Response) -> None:
        nonlocal call_count
        call_count += 1
        if call_count < 2:
            raise httpx.TimeoutException("not ready yet")

    retry = Retry(total=5, validate_response=validate)
    transport = RetryTransport(retry=retry)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")

    assert response.status_code == 200
    assert call_count == 2
    assert mock_asleep.call_count == 1


@pytest.mark.asyncio
async def test_async_validate_response_exhausted_returns_response(mock_async_responses: AsyncMockResponse) -> None:
    mock_asleep, _ = mock_async_responses

    async def validate(response: httpx.Response) -> None:
        raise httpx.TimeoutException("always bad")

    retry = Retry(total=3, validate_response=validate)
    transport = RetryTransport(retry=retry)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("https://example.com")

    assert response.status_code == 200
    assert mock_asleep.call_count == 3


def test_validate_response_not_called_for_retryable_status(mock_responses: MockResponse) -> None:
    mock_sleep, status_code_sequences = mock_responses
    status_code_sequences["https://example.com/fail"] = status_codes([(503, None), (200, None)])
    validated = []

    def validate(response: httpx.Response) -> None:
        validated.append(response.status_code)

    retry = Retry(total=5, validate_response=validate)
    transport = RetryTransport(retry=retry)

    with httpx.Client(transport=transport) as client:
        response = client.get("https://example.com/fail")

    assert response.status_code == 200
    assert validated == [200]



import datetime
import inspect
import logging
from http import HTTPStatus
from typing import Any
from unittest.mock import AsyncMock, MagicMock

import httpx
import pytest
from httpx import Headers, Response

from httpx_retries.retry import HTTPMethod, Retry


def test_retry_initialization() -> None:
    retry = Retry()
    assert retry.total == 10
    assert retry.backoff_factor == 0
    assert retry.respect_retry_after_header is True
    assert retry.max_backoff_wait == 120


def test_retry_custom_initialization() -> None:
    retry = Retry(
        total=5,
        backoff_factor=0.5,
        respect_retry_after_header=False,
        max_backoff_wait=30,
        allowed_methods=["GET", "POST"],
        status_forcelist=[500, 502],
    )
    assert retry.total == 5
    assert retry.backoff_factor == 0.5
    assert retry.respect_retry_after_header is False
    assert retry.max_backoff_wait == 30
    assert HTTPMethod.GET in retry.allowed_methods
    assert HTTPMethod.POST in retry.allowed_methods
    assert HTTPStatus.INTERNAL_SERVER_ERROR in retry.status_forcelist
    assert HTTPStatus.BAD_GATEWAY in retry.status_forcelist


def test_is_retryable_method() -> None:
    retry = Retry()
    assert retry.is_retryable_method("GET") is True
    assert retry.is_retryable_method("POST") is False


def test_is_retryable_status_code() -> None:
    retry = Retry()
    assert retry.is_retryable_status_code(429) is True
    assert retry.is_retryable_status_code(404) is False


def test_is_retryable_exception() -> None:
    retry = Retry()
    assert retry.is_retryable_exception(httpx.NetworkError("")) is True
    assert retry.is_retryable_exception(httpx.LocalProtocolError("")) is False


def test_is_retryable_exception_custom_exception() -> None:
    class MyExc(Exception):
        pass

    retry = Retry(retry_on_exceptions=(MyExc,))
    assert retry.is_retryable_exception(httpx.NetworkError("")) is False
    assert retry.is_retryable_exception(MyExc()) is True


def test_custom_retryable_methods_str() -> None:
    retry = Retry(allowed_methods=["POST"])
    assert retry.is_retryable_method("POST") is True
    assert retry.is_retryable_method("GET") is False


def test_custom_retryable_methods_enum() -> None:
    retry = Retry(allowed_methods=[HTTPMethod.POST])
    assert retry.is_retryable_method("POST") is True
    assert retry.is_retryable_method("GET") is False


def test_non_standard_method_not_retryable_by_default() -> None:
    retry = Retry()
    assert retry.is_retryable_method("PROPFIND") is False
    assert retry.is_retryable_method("propfind") is False


def test_non_standard_method_can_be_configured() -> None:
    retry = Retry(allowed_methods=["PROPFIND", "GET"])
    assert retry.is_retryable_method("PROPFIND") is True
    assert retry.is_retryable_method("propfind") is True
    assert retry.is_retryable_method("GET") is True
    assert retry.is_retryable_method("POST") is False


def test_custom_retry_status_codes() -> None:
    retry = Retry(status_forcelist=[500])
    assert retry.is_retryable_status_code(500) is True
    assert retry.is_retryable_status_code(502) is False


def test_custom_retry_status_codes_enum() -> None:
    retry = Retry(status_forcelist=[HTTPStatus.INTERNAL_SERVER_ERROR])
    assert retry.is_retryable_status_code(500) is True
    assert retry.is_retryable_status_code(502) is False


def test_custom_retry_status_codes_non_standard() -> None:
    retry = Retry(status_forcelist=[523, 599])
    assert retry.is_retryable_status_code(523) is True
    assert retry.is_retryable_status_code(599) is True
    assert retry.is_retryable_status_code(500) is False
    assert retry.is_retryable_status_code(502) is False


def test_is_exhausted() -> None:
    retry = Retry(total=3)
    assert retry.is_exhausted() is False

    retry = retry.increment().increment().increment()
    assert retry.is_exhausted() is True


def test_parse_retry_after_seconds() -> None:
    retry = Retry()
    assert retry.parse_retry_after("5") == 5.0


def test_parse_retry_after_http_date() -> None:
    retry = Retry()
    future_date = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=5)).strftime(
        "%a, %d %b %Y %H:%M:%S GMT"
    )
    result = retry.parse_retry_after(future_date)
    assert 3 < result < 7  # Allow some flexibility for test execution time


def test_parse_retry_after_http_date_past() -> None:
    retry = Retry()
    past_date = (datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(seconds=5)).strftime(
        "%a, %d %b %Y %H:%M:%S GMT"
    )
    result = retry.parse_retry_after(past_date)
    assert result == 0


def test_parse_retry_after_http_date_no_tz() -> None:
    retry = Retry()
    future_date = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=5)).strftime(
        "%a, %d %b %Y %H:%M:%S"
    )
    result = retry.parse_retry_after(future_date)
    assert 3 < result < 7


def test_parse_retry_after_http_date_no_tz_logs_warning(caplog: pytest.LogCaptureFixture) -> None:
    import logging

    retry = Retry()
    future_date = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=5)).strftime(
        "%a, %d %b %Y %H:%M:%S"
    )
    with caplog.at_level(logging.WARNING, logger="httpx_retries.retry"):
        retry.parse_retry_after(future_date)
    assert "assuming UTC" in caplog.text


def test_calculate_sleep_with_retry_after_over_max() -> None:
    retry = Retry(max_backoff_wait=5)
    headers = Headers({"Retry-After": "10"})
    assert retry._calculate_sleep(headers) == 5.0


def test_calculate_sleep_respect_retry_after_false() -> None:
    retry = Retry(respect_retry_after_header=False)
    headers = Headers({"Retry-After": "5"})
    sleep_time = retry._calculate_sleep(headers)
    assert sleep_time <= retry.max_backoff_wait


def test_parse_retry_after_invalid() -> None:
    retry = Retry()
    with pytest.raises(ValueError):
        retry.parse_retry_after("invalid")


@pytest.mark.parametrize("value", ["\u0665", "\u00b2"])
def test_parse_retry_after_rejects_non_ascii_digits(value: str) -> None:
    retry = Retry()
    with pytest.raises(ValueError, match=f"Invalid Retry-After header: {value}"):
        retry.parse_retry_after(value)


def test_calculate_sleep_with_retry_after() -> None:
    retry = Retry()
    headers = Headers({"Retry-After": "5"})
    assert retry._calculate_sleep(headers) == 5.0


def test_calculate_sleep_with_invalid_retry_after(caplog: pytest.LogCaptureFixture) -> None:
    retry = Retry()
    headers = Headers({"Retry-After": "invalid"})
    sleep_time = retry._calculate_sleep(headers)

    assert sleep_time <= retry.max_backoff_wait
    assert "Retry-After header is not a valid HTTP date" in caplog.text


def test_calculate_sleep_returns_immediately_on_first_attempt() -> None:
    retry = Retry()
    headers = Headers({})
    assert retry._calculate_sleep(headers) == 0


def test_sleep_respects_retry_after_header(mock_sleep: MagicMock) -> None:
    retry = Retry()
    response = Response(status_code=429, headers={"Retry-After": "5"})
    retry.sleep(response)
    assert mock_sleep.call_count == 1
    mock_sleep.assert_called_with(5.0)


def test_sleep_logs_sleep_time(mock_sleep: MagicMock, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG)
    retry = Retry()
    response = Response(status_code=429, headers={"Retry-After": "5"})
    retry.sleep(response)
    assert "sleep seconds=5.0" in caplog.text


@pytest.mark.asyncio
async def test_asleep_respects_retry_after_header(mock_asleep: AsyncMock) -> None:
    retry = Retry()
    response = Response(status_code=429, headers={"Retry-After": "5"})
    await retry.asleep(response)
    assert mock_asleep.call_count == 1
    mock_asleep.assert_called_with(5.0)


@pytest.mark.asyncio
async def test_asleep_logs_sleep_time(mock_asleep: AsyncMock, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG)
    retry = Retry()
    response = Response(status_code=429, headers={"Retry-After": "5"})
    await retry.asleep(response)
    assert "asleep seconds=5.0" in caplog.text


def test_calculate_sleep_returns_immediately_by_default() -> None:
    # If there is no backoff factor or retry after header, the sleep time should be 0
    retry = Retry(attempts_made=1)
    headers = Headers({})
    assert retry._calculate_sleep(headers) == 0


def test_calculate_sleep_with_backoff() -> None:
    retry = Retry(backoff_factor=2)
    headers = Headers({})
    attempts: list[Retry] = [retry]
    for _ in range(3):
        attempts.append(attempts[-1].increment())

    for i, retry_attempt in enumerate(attempts):
        sleep_time = retry_attempt._calculate_sleep(headers)
        # Backoff should be backoff_factor * (2 ** (attempts))
        # With full jitter, the actual sleep time will be between 0 and this value
        max_expected = 2 * (2 ** (i))
        assert sleep_time <= max_expected


def test_calculate_sleep_falls_back_to_backoff_if_retry_after_is_in_the_past() -> None:
    retry = Retry(backoff_factor=2, attempts_made=2)
    headers = Headers({"Retry-After": "0"})
    sleep_time = retry._calculate_sleep(headers)
    assert sleep_time != 0
    assert sleep_time <= 2 * (2 ** (2))


def test_calculate_sleep_max_backoff() -> None:
    retry = Retry(backoff_factor=2, max_backoff_wait=5)
    headers = Headers({})
    # After several attempts, backoff would exceed max_backoff_wait
    retry = retry.increment().increment().increment()
    sleep_time = retry._calculate_sleep(headers)
    assert sleep_time <= 5


def test_calculate_sleep_first_attempt() -> None:
    retry = Retry(backoff_factor=2)
    headers = Headers({})
    sleep_time = retry._calculate_sleep(headers)
    # First attempt should have no backoff
    assert sleep_time <= 2


def test_calculate_sleep_without_retry_after() -> None:
    retry = Retry()
    headers = Headers({})
    sleep_time = retry._calculate_sleep(headers)

    assert sleep_time <= retry.max_backoff_wait


def test_increment() -> None:
    retry = Retry(total=3)
    new_retry = retry.increment()

    assert new_retry != retry

    assert new_retry.attempts_made == retry.attempts_made + 1
    assert new_retry.total == retry.total
    assert new_retry.backoff_factor == retry.backoff_factor
    assert new_retry.respect_retry_after_header == retry.respect_retry_after_header
    assert new_retry.allowed_methods == retry.allowed_methods
    assert new_retry.status_forcelist == retry.status_forcelist


def test_increment_logs(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG)
    retry = Retry(total=3)
    new_retry = retry.increment()
    assert "increment retry=<Retry(total=3, attempts_made=0)> new_attempts_made=1" in caplog.text
    new_retry.increment()
    assert "increment retry=<Retry(total=3, attempts_made=1)> new_attempts_made=2" in caplog.text


def test_retry_validation_negative_total() -> None:
    with pytest.raises(ValueError, match="total must be non-negative"):
        Retry(total=-1)


def test_retry_validation_negative_backoff_factor() -> None:
    with pytest.raises(ValueError, match="backoff_factor must be non-negative"):
        Retry(backoff_factor=-0.5)


def test_retry_validation_zero_max_backoff_wait() -> None:
    with pytest.raises(ValueError, match="max_backoff_wait must be positive"):
        Retry(max_backoff_wait=0)


def test_retry_validation_negative_max_backoff_wait() -> None:
    with pytest.raises(ValueError, match="max_backoff_wait must be positive"):
        Retry(max_backoff_wait=-1)


def test_retry_validation_invalid_jitter() -> None:
    with pytest.raises(ValueError, match="backoff_jitter must be between 0 and 1"):
        Retry(backoff_jitter=1.5)


def test_retry_validation_negative_jitter() -> None:
    with pytest.raises(ValueError, match="backoff_jitter must be between 0 and 1"):
        Retry(backoff_jitter=-0.5)


def test_retry_validation_negative_attempts() -> None:
    with pytest.raises(ValueError, match="attempts_made must be non-negative"):
        Retry(attempts_made=-1)


def test_method_case_insensitive() -> None:
    retry = Retry(allowed_methods=["get", "POST"])
    assert retry.is_retryable_method("GET")
    assert retry.is_retryable_method("get")
    assert retry.is_retryable_method("POST")
    assert retry.is_retryable_method("post")


def test_backoff_with_no_jitter() -> None:
    retry = Retry(backoff_factor=1, backoff_jitter=0)
    retry = retry.increment()  # One attempt made

    # With no jitter, backoff should be exactly backoff_factor * (2 ** attempts_made)
    assert retry.backoff_strategy() == 2.0


def test_backoff_with_partial_jitter() -> None:
    retry = Retry(backoff_factor=1, backoff_jitter=0.5)
    retry = retry.increment()  # One attempt made

    # With 0.5 jitter, backoff should be between 1.0 and 2.0 times backoff_factor * (2 ** attempts_made)
    backoff = retry.backoff_strategy()
    assert 1.0 <= backoff <= 2.0


def test_zero_backoff_factor() -> None:
    retry = Retry(backoff_factor=0)
    retry = retry.increment()

    assert retry.backoff_strategy() == 0.0


def test_increment_preserves_jitter() -> None:
    retry = Retry(backoff_jitter=0.5)
    new_retry = retry.increment()

    assert new_retry.backoff_jitter == retry.backoff_jitter


def test_retry_after_header_invalid_format() -> None:
    retry = Retry()
    with pytest.raises(ValueError, match="Invalid Retry-After header: invalid date"):
        retry.parse_retry_after("invalid date")


def test_retry_after_header_logging(caplog: pytest.LogCaptureFixture) -> None:
    retry = Retry()
    headers = Headers({"Retry-After": "invalid date"})
    retry._calculate_sleep(headers)

    assert "Retry-After header is not a valid HTTP date: invalid date" in caplog.text


def test_retry_after_header_precedence() -> None:
    retry = Retry(backoff_factor=2)
    retry = retry.increment()  # One attempt made

    # Retry-After header should take precedence over backoff strategy
    headers = Headers({"Retry-After": "1"})
    sleep_time = retry._calculate_sleep(headers)

    assert sleep_time == 1.0


def test_retry_after_respects_max_wait() -> None:
    retry = Retry(max_backoff_wait=1)
    headers = Headers({"Retry-After": "10"})
    sleep_time = retry._calculate_sleep(headers)

    assert sleep_time == 1.0


def test_default_values() -> None:
    retry = Retry()
    assert retry.total == 10
    assert retry.backoff_factor == 0.0
    assert retry.max_backoff_wait == 120.0
    assert retry.backoff_jitter == 1.0


def test_is_retry() -> None:
    retry = Retry(total=3)
    assert retry.is_retry("GET", 429, False) is True
    assert retry.is_retry("POST", 429, False) is False
    assert retry.is_retry("GET", 404, False) is False
    assert retry.is_retry("GET", 429, True) is False

    exhausted_retry = Retry(total=0)
    assert exhausted_retry.is_retry("GET", 429, False) is False


def test_is_retry_custom_configuration() -> None:
    retry = Retry(total=3, allowed_methods=["POST"], status_forcelist=[404])
    assert retry.is_retry("POST", 404, False) is True
    assert retry.is_retry("GET", 404, False) is False
    assert retry.is_retry("POST", 429, False) is False


def test_total_timeout_default_is_none() -> None:
    retry = Retry()
    assert retry.total_timeout is None
    assert retry.elapsed_sleep == 0.0


def test_total_timeout_custom() -> None:
    retry = Retry(total_timeout=30.0)
    assert retry.total_timeout == 30.0


def test_retry_validation_negative_total_timeout() -> None:
    with pytest.raises(ValueError, match="total_timeout must be positive"):
        Retry(total_timeout=-1)


def test_retry_validation_zero_total_timeout() -> None:
    with pytest.raises(ValueError, match="total_timeout must be positive"):
        Retry(total_timeout=0)


def test_retry_validation_negative_elapsed_sleep() -> None:
    with pytest.raises(ValueError, match="elapsed_sleep must be non-negative"):
        Retry(elapsed_sleep=-0.1)


def test_calculate_sleep_capped_by_total_timeout() -> None:
    retry = Retry(total_timeout=5)
    headers = Headers({"Retry-After": "10"})
    assert retry._calculate_sleep(headers) == 5.0


def test_calculate_sleep_capped_by_remaining_total_timeout() -> None:
    retry = Retry(total_timeout=10)
    retry.elapsed_sleep = 7.0
    headers = Headers({"Retry-After": "10"})
    assert retry._calculate_sleep(headers) == 3.0


def test_calculate_sleep_zero_when_total_timeout_exhausted() -> None:
    retry = Retry(total_timeout=5)
    retry.elapsed_sleep = 5.0
    headers = Headers({"Retry-After": "10"})
    assert retry._calculate_sleep(headers) == 0.0


def test_calculate_sleep_no_cap_without_total_timeout() -> None:
    retry = Retry()
    headers = Headers({"Retry-After": "10"})
    assert retry._calculate_sleep(headers) == 10.0


def test_sleep_updates_elapsed_sleep(mock_sleep: MagicMock) -> None:
    retry = Retry()
    assert retry.elapsed_sleep == 0.0
    response = Response(status_code=429, headers={"Retry-After": "5"})
    retry.sleep(response)
    assert retry.elapsed_sleep == 5.0
    retry.sleep(response)
    assert retry.elapsed_sleep == 10.0


@pytest.mark.asyncio
async def test_asleep_updates_elapsed_sleep(mock_asleep: AsyncMock) -> None:
    retry = Retry()
    assert retry.elapsed_sleep == 0.0
    response = Response(status_code=429, headers={"Retry-After": "5"})
    await retry.asleep(response)
    assert retry.elapsed_sleep == 5.0
    await retry.asleep(response)
    assert retry.elapsed_sleep == 10.0


def test_increment_preserves_total_timeout() -> None:
    retry = Retry(total_timeout=30.0)
    new_retry = retry.increment()
    assert new_retry.total_timeout == 30.0


def test_increment_preserves_elapsed_sleep() -> None:
    retry = Retry(total_timeout=30.0)
    retry.elapsed_sleep = 12.5
    new_retry = retry.increment()
    assert new_retry.elapsed_sleep == 12.5


def test_is_exhausted_by_total_timeout() -> None:
    retry = Retry(total=10, total_timeout=5)
    retry.elapsed_sleep = 5.0
    assert retry.is_exhausted() is True


def test_is_exhausted_below_total_timeout() -> None:
    retry = Retry(total=10, total_timeout=5)
    retry.elapsed_sleep = 4.9
    assert retry.is_exhausted() is False


def test_copy_with_overrides_fields() -> None:
    retry = Retry(total=10, backoff_factor=0.5, total_timeout=30.0)
    copy = retry.copy_with(total=3, backoff_factor=1.0)
    assert copy.total == 3
    assert copy.backoff_factor == 1.0
    assert copy.total_timeout == 30.0  # unchanged


def test_copy_with_no_args_equals_original() -> None:
    retry = Retry(total=5, backoff_factor=0.2, max_backoff_wait=60.0)
    copy = retry.copy_with()
    assert copy.total == retry.total
    assert copy.backoff_factor == retry.backoff_factor
    assert copy.max_backoff_wait == retry.max_backoff_wait
    assert copy.allowed_methods == retry.allowed_methods
    assert copy.status_forcelist == retry.status_forcelist
    assert copy.retryable_exceptions == retry.retryable_exceptions


def test_copy_with_can_set_total_timeout_to_none() -> None:
    retry = Retry(total=5, total_timeout=10.0)
    copy = retry.copy_with(total_timeout=None)
    assert copy.total_timeout is None


def test_copy_with_preserves_subclass() -> None:
    class CustomRetry(Retry):
        pass

    retry = CustomRetry(total=5)
    copy = retry.copy_with(total=3)
    assert type(copy) is CustomRetry


def test_copy_with_and_init_have_same_parameters() -> None:
    init_params = set(inspect.signature(Retry.__init__).parameters) - {"self"}
    copy_with_params = set(inspect.signature(Retry.copy_with).parameters) - {"self"}
    assert init_params == copy_with_params


def test_copy_with_roundtrips_all_fields() -> None:
    init_params = set(inspect.signature(Retry.__init__).parameters) - {"self"}
    kwargs: dict[str, Any] = {
        "total": 3,
        "allowed_methods": {"GET", "POST"},
        "status_forcelist": {500, 503},
        "retry_on_exceptions": [httpx.TimeoutException],
        "backoff_factor": 0.5,
        "respect_retry_after_header": False,
        "max_backoff_wait": 60.0,
        "backoff_jitter": 0.5,
        "attempts_made": 2,
        "total_timeout": 30.0,
        "elapsed_sleep": 1.5,
        "validate_response": lambda _: None,
    }
    assert set(kwargs) == init_params

    original = Retry(**kwargs)
    copy = original.copy_with(**kwargs)

    for attr, value in vars(original).items():
        assert getattr(copy, attr) == value, f"{attr} mismatch"
